"""
This uses Paramiko to send files from workers to other workers (without having
to go through the master).
"""

from reductio import para_ssh
from reductio import config
import logging
import os
LOG = logging.getLogger(__name__)

def send_file(connection, source_path, target_path):
    """
    `source_path` is the absolute path of the file to send.
    `target_path` is the relative path of where to put it.
    """
    absolute_target = os.path.join(connection.reductio_home, target)
    target_dir = os.path.dirname(absolute_target)
    connection.execute("mkdir -p %s" % target_dir)
    connection.put(source_path, absolute_target)

def make_connections():
    """
    Get an ssh.Connection object for each worker that we can successfully
    contact and run reductio on.
    """
    connections = []
    for worker_config in config.WORKERS:
        conn = ssh.Connection(**worker_config)
        if worker_config['virtualenv']:
            conn.execute("source %s/bin/activate" % worker_config['virtualenv'])
        try:
            output = conn.execute("python -m reductio.config HOMEDIR")[0].strip()
        except:
            log.warn("Could not connect to %r" % worker_config)
            continue
        conn.reductio_home = output[0].strip()
        connections.append(conn)
    if len(connections) == 0:
        raise IOError("reductio could not connect to any workers.")
    return connections

CONNECTIONS = None

def get_connections():
    """
    Get the list of connections if it already exists; create it using
    `make_connections` if not.
    """
    global CONNECTIONS
    if CONNECTIONS is None:
        CONNECTIONS = make_connections()
    return CONNECTIONS

def close_connections():
    if CONNECTIONS is not None:
        for conn in CONNECTIONS:
            conn.close()
    CONNECTIONS = None

def __del__():
    close_connections()




from fabric.api import env, task, run, sudo, put, serial
import config
import os

env.parallel = True
env.hosts = config.FABRIC_HOSTS

def map(function, source, target):
    run('python -m reductio.operate map %s %s %s' % (function, source, target))

def reduce(function, source, target):
    run('python -m reductio.operate reduce %s %s %s' % (function, source, target))

def transform(function, source, target):
    run('python -m reductio.operate transform %s %s %s' % (function, source, target))

def initialize(function, target):
    run('python -m reductio.operate initialize %s %s %s' % (function, target, target))

def scatter(source, target):
    run('python -m reductio.operate scatter %s %s' % (source, target))

def sort(source, target):
    run('python -m reductio.operate sort %s %s' % (source, target))

def sort_unique(source, target):
    run('python -m reductio.operate sort_unique %s %s' % (source, target))

def delete(source):
    run('python -m reductio.operate delete %s %s' % (source, source))

# TODO: account for virtualenvs defined in config
@serial
def install_package(package_name):
    run('source ~/.bashrc')
    env = run('echo $VIRTUAL_ENV')
    if env.strip():
        run('easy_install -U %s' % package_name)
    else:
        sudo('easy_install -U %s' % package_name)

@serial
@task
def install_reductio():
    install_package('reductio')

@serial
@task
def configure_ubuntu():
    uname = run('uname -a')
    if 'Ubuntu' in uname:
        sudo('aptitude update')
        sudo('aptitude install -y python-setuptools python-dev build-essential')

@task
def get_worker_home():
    return run('python -m reductio.config HOMEDIR').strip()

def put_in_directory(source_path, target_path):
    target_path = str(run('echo %s' % target_path))
    run('mkdir -p '+os.path.dirname(target_path))
    if env.host != config.HOSTNAME or source_path != target_path:
        put(source_path, target_path)
    else:
        print "Skipping identical file"

# TODO: we might not keep the config in HOMEDIR
@task
def deploy_worker_config(dir=None):
    if dir is None:
        dir = '~/.reductio'
    source_file = os.path.join(config.HOMEDIR, 'workers.cfg')
    target_file = os.path.join(dir, 'workers.cfg')
    put_in_directory(source_file, target_file)




import os
import sys
import socket
import json
import ConfigParser

HOSTNAME = socket.gethostname()
HOMEDIR = os.path.expanduser('~/.reductio')
PARTITIONS = 64

def get_worker_config(dirlist):
    worker_config = ConfigParser.ConfigParser()
    file_possibilities = [os.path.join(dir, 'workers.cfg') for dir in dirlist]
    got = worker_config.read(file_possibilities)
    workers = []
    for section in worker_config.sections():
        worker_dict = {}
        for key, value in worker_config.items(section):
            worker_dict[key] = value
        workers.append(worker_dict)
    return workers

def worker_to_fabric(worker):
    if 'username' in worker:
        hoststr = worker['username'] + '@' + worker['host']
    else:
        hoststr = worker['host']
    if 'password' in worker:
        from fabric.api import env
        env.passwords[hoststr] = worker['password']
    return hoststr

WORKERS = get_worker_config(['.', HOMEDIR])
FABRIC_HOSTS = [worker_to_fabric(w) for w in WORKERS]
#global_config = ConfigParser.ConfigParser()
#global_config.read([os.path.join(HOMEDIR, 'reductio.cfg')])

def main(command, *args):
    if command in globals():
        print globals()[command]
    else:
        raise ValueError("There is no configuration option called %s.\n" % command)

if __name__ == '__main__':
    if len(sys.argv) <= 1:
        sys.stderr.write("Usage: python -m reductio.config (command)\n")
    main(*sys.argv[1:])



import subprocess

def sort(filename_in, filename_out, unique=False):
    if unique:
        subprocess.check_call(['sort', filename_in, '-u', '-o', filename_out])
    else:
        subprocess.check_call(['sort', filename_in, '-o', filename_out])



from reductio import config
from reductio import network
from reductio import process
import itertools
import sys
import os
import shutil
from importlib import import_module

def _exists(filename):
    "Test whether a file exists."
    return os.access(filename, os.F_OK)

class ShardFileStorage(object):
    """
    Reads and writes data in sharded files that can be distributed.

    This takes output from a map or reduce function, and writes it
    into outgoing files that will be sharded by key and distributed to
    the different workers.
    
    If the output files already exist, it will append to them, so that merging
    data sets is easy.

    This can be written to as a file-like object. Data written to it in this
    way will go to the unformatted text file "result.txt" in its directory,
    which can be used as final output or passed to a reducer on the same
    machine.
    
    All text is read and written as UTF-8.
    """
    def __init__(self, input_name, output_name,
                 homedir=config.HOMEDIR,
                 partitions=config.PARTITIONS):
        self.partitions = partitions
        self.homedir = homedir
        self.input_dir = homedir + os.path.sep + input_name
        self.output_name = output_name
        self.output_dir = homedir + os.path.sep + output_name
        for dir in (self.input_dir, self.output_dir):
            if not _exists(dir):
                os.makedirs(dir)
                print "Created directory %r" % dir

        self.output_files = [None] * self.partitions
        self.result_file = None
    
    def __iter__(self):
        return self.iterlines()

    def input_files(self):
        """
        Get the filenames for files that are inputs to this operation."
        """
        for filename in os.listdir(self.input_dir):
            if filename.endswith('.txt'):
                yield os.path.join(self.input_dir, filename)

    def iterlines(self):
        """
        Returns an iterator over lines in the input files. Skips blank lines.
        """
        for filename in self.input_files():
            infile = open(filename)
            for line in infile:
                if line.strip('\n'):
                    yield line.decode('utf-8')

    def get_partition(self, key):
        return hash(key) % self.partitions

    def write_to_file(self, index, data):
        if self.output_files[index] is None:
            filename = self.output_dir + os.path.sep + "%d.txt" % index
            self.output_files[index] = open(filename, 'a')
        self.output_files[index].write(data)

    def write_pair(self, key, value):
        """
        Write a key-value pair to the appropriate output file, so later it
        can be sent to another worker.
        """
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        partition = self.get_partition(key)
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        self.write_to_file(partition, "%s\t%s\n" % (key, value))

    def write(self, data):
        """
        Write unstructured data to the file `result.txt`.
        """
        if self.result_file is None:
            filename = self.output_dir + os.path.sep + 'result.txt'
            self.result_file = open(filename, 'a')
        if isinstance(data, unicode):
            data = data.encode('utf-8')
        self.result_file.write(data)

    def do_scatter(self):
        """
        Distributes the output files to other servers, where they will
        become input files.
        """
        connections = network.get_connections()
        for (filename, connection) in zip(
            self.input_files(),
            itertools.cycle(connections)
        ):
            source_path = os.path.join(self.input_dir, filename)
            target_path = os.path.join(
                self.output_name,
                config.HOSTNAME+'_'+filename
            )
            network.send_file(connection, source_path, target_path)

    def do_sort(self):
        for file in self.input_files():
            target_file = os.path.join(self.output_dir, os.path.basename(file))
            process.sort(file, target_file)
    
    def do_sort_unique(self):
        for file in self.input_files():
            target_file = os.path.join(self.output_dir, os.path.basename(file))
            process.sort(file, target_file, unique=True)

    def do_initialize(self, input_generator):
        input_generator(self)

    def do_delete(self):
        target_dir = os.path.normpath(self.input_dir)
        if not target_dir.startswith(self.homedir):
            raise ValueError(
                "%r doesn't seem to be a subdirectory of %r. "
                "Refusing to delete it, to be safe."
                % (target_dir, self.homedir)
            )
        shutil.rmtree(target_dir)

    def do_transform(self, transformer):
        r"""
        A "transform" is a kind of map operation that makes no assumptions about
        its data other than the fact that it's Unicode text.
        
        The cases where you need a transform are when

        - your input data isn't in reductio's key-value form, or
        - you want to produce output that isn't in reductio's key-value form.

        As such, a transform function gets the raw text of each line, plus a
        file-like object that you can write to (or output key/value pairs to,
        because it's in fact this object).
        
        If you want to use this data in future
        steps of reductio, your output should be lines in the form:

            <key>\t<value>\n
        
        You are not required to produce one line of output per line of input.
        """
        for line in input_stream:
            transformer(line, self)

    def do_map(self, mapper):
        """
        This function makes it convenient to write a pure mapping function, as
        a generator that takes in (key, value) pairs and yields (key, value)
        pairs.
        """
        for line in self:
            key, value = line.strip('\n').split('\t', 1)
            for result_key, result_value in mapper(key, value):
                self.write_pair(result_key, result_value)
    
    def do_reduce(self, reducer):
        """
        A reducer differs from a mapper in that it groups its incoming data by
        key, and gives the reduce function a generator of the associated values.

        Unlike some map-reduce frameworks, this will never decide to reduce in
        multiple stages. It just gives you an iterator and assumes you can
        handle things from there.
        
        This means your reduce function does *not* have to be prepared to
        take its own output as input. Try taking advantage of this by performing
        your next "map" operation right here!
        """
        for key, pairs in itertools.groupby(
            pair_iterator(self),
            _key_selector
        ):
            values = (_value_selector(pair) for pair in pairs)
            for out_key, out_value in reducer(key, values):
                self.write_pair(out_key, out_value)

def pair_iterator(input_stream):
    """
    Extract tab-delimited (key, value) pairs from a file.
    """
    for line in input_stream:
        line = input_stream.next().strip('\n')
        if line:
            key, value = line.split('\t', 1)
            yield key, value

def _key_selector(pair):
    return pair[0]

def _value_selector(pair):
    return pair[1]

def run_scatter(source, target):
    """
    Send output files to other workers, where they will become input files.
    """
    storage = ShardFileStorage(source, target)
    storage.do_scatter()

def run(command, *args):
    name = 'do_'+command
    if command in ('scatter', 'sort', 'sort_unique', 'delete'):
        if len(args) != 2:
            raise ValueError(
                "Wrong number of arguments. "
                "Expected (operation, source, target)."
        )
        [source, target] = args
        storage = ShardFileStorage(source, target)
        operator = getattr(storage, name)
        operator()
    else:
        if len(args) != 3:
            raise ValueError(
                "Wrong number of arguments. "
                "Expected (operation, function, source, target). "
                "Got: %r" % ([command]+list(args))
            )
        function_path, source, target = args
        module_name, function_name = function_path.rsplit('.', 1)
        storage = ShardFileStorage(source, target)
        if hasattr(storage, name):
            module = import_module(module_name)
            func = getattr(module, function_name)
            operator = getattr(storage, name)
            if func is None:
                raise ValueError("Cannot find %r in %r." % (function_name, module))
            operator(func)
        else:
            raise ValueError("There is no command called %r.\n" % command)

if __name__ == '__main__':
    run(*sys.argv[1:])




"""
This code is derived from Zeth's "Friendly Python SSH2 interface":
http://media.commandline.org.uk/code/ssh.txt
"""

import os
import tempfile
import paramiko

class Connection(object):
    """Connects and logs into the specified hostname. 
    Arguments that are not given are guessed from the environment.""" 

    def __init__(self,
                 host,
                 username = None,
                 private_key = None,
                 password = None,
                 port = 22,
                 ):
        self._sftp_live = False
        self._sftp = None
        self.host = host
        if not username:
            username = os.environ['LOGNAME']

        # Log to a temporary file.
        templog = tempfile.mkstemp('.txt', 'ssh-')[1]
        paramiko.util.log_to_file(templog)

        # Begin the SSH transport.
        self._transport = paramiko.Transport((host, port))
        self._tranport_live = True
        # Authenticate the transport.
        if password:
            # Using Password.
            self._transport.connect(username = username, password = password)
        else:
            # Use Private Key.
            if not private_key:
                # Try to use default key.
                if os.path.exists(os.path.expanduser('~/.ssh/id_rsa')):
                    private_key = '~/.ssh/id_rsa'
                elif os.path.exists(os.path.expanduser('~/.ssh/id_dsa')):
                    private_key = '~/.ssh/id_dsa'
                else:
                    raise TypeError, "You have not specified a password or key."

            private_key_file = os.path.expanduser(private_key)
            rsa_key = paramiko.RSAKey.from_private_key_file(private_key_file)
            self._transport.connect(username = username, pkey = rsa_key)
    
    def _sftp_connect(self):
        """Establish the SFTP connection."""
        if not self._sftp_live:
            self._sftp = paramiko.SFTPClient.from_transport(self._transport)
            self._sftp_live = True

    def get(self, remotepath, localpath = None):
        """Copies a file between the remote host and the local host."""
        if not localpath:
            localpath = os.path.split(remotepath)[1]
        self._sftp_connect()
        self._sftp.get(remotepath, localpath)

    def put(self, localpath, remotepath = None):
        """Copies a file between the local host and the remote host."""
        if not remotepath:
            remotepath = os.path.split(localpath)[1]
        self._sftp_connect()
        self._sftp.put(localpath, remotepath)

    def execute(self, command):
        """Execute the given commands on a remote machine."""
        channel = self._transport.open_session()
        channel.exec_command(command)
        output = channel.makefile('rb', -1).readlines()
        if output:
            return output
        else:
            return channel.makefile_stderr('rb', -1).readlines()

    def close(self):
        """Closes the connection and cleans up."""
        # Close SFTP Connection.
        if self._sftp_live:
            self._sftp.close()
            self._sftp_live = False
        # Close the SSH Transport.
        if self._tranport_live:
            self._transport.close()
            self._tranport_live = False

    def __repr__(self):
        return "<Connection: %s>" % self.host

    def __del__(self):
        """Attempt to clean up if not explicitly closed."""
        self.close()

def main():
    """Little test when called directly."""
    # Set these to your own details.
    myssh = Connection('example.com')
    myssh.put('ssh.py')
    myssh.close()

# start the ball rolling.
if __name__ == "__main__":
    main()




import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "reductio",
    version = "0.1.4",
    author = "Rob Speer",
    author_email = "rspeer@mit.edu",
    description = ("Map/Reduce for Fabric"),
    license = "AGPLv3",
    keywords = "mapreduce distributed fabric",
    url = "http://github.com/rspeer/reductio",
    packages=['reductio'],
    install_requires=['fabric', 'paramiko', 'importlib'],
    long_description=read('README'),
)
